@@ -75,7 +75,6 @@
 #include "nsIWebBrowser.h"
 #include "nsIWebNavigation.h"
 
-#include "gfxIImageFrame.h"
 // imglib2
 #include "imgIRequest.h"
 #include "imgIContainer.h"
@@ -128,7 +127,7 @@
 //Scrolling internal frames
 #include "nsIContent.h"
 #include "nsQueryFrame.h"
-#include "nsIFrame.h"
+//#include "nsIFrame.h"
 #include "nsIScrollableView.h"
 #include "nsIScrollableViewProvider.h"
 #include "nsIView.h"
@@ -142,6 +141,8 @@
 #include "nsIDOMXULTextboxElement.h"
 #endif
 
+#include "nsIHTMLDocument.h"
+
 //*****************************************************************************
 // class MicrobEalContextMenuInfo
 //*****************************************************************************
@@ -328,6 +329,8 @@ MicrobEalContextMenuInfo::CheckFormControl(nsIDOMNode* aNode)
   {
     nsCOMPtr<nsIDOMHTMLButtonElement> disForm = do_QueryInterface(tgContent);
     if (disForm) disForm->GetDisabled(&isDisabled);
+    clickableForm = PR_TRUE;
+    break;
   }
   case NS_FORM_INPUT_BUTTON:
   case NS_FORM_INPUT_CHECKBOX:
@@ -338,11 +341,15 @@ MicrobEalContextMenuInfo::CheckFormControl(nsIDOMNode* aNode)
   {
     nsCOMPtr<nsIDOMHTMLInputElement> disForm = do_QueryInterface(tgContent);
     if (disForm) disForm->GetDisabled(&isDisabled);
+    clickableForm = PR_TRUE;
+    break;
   }
   case NS_FORM_OPTION:
   {
     nsCOMPtr<nsIDOMHTMLOptionElement> disForm = do_QueryInterface(tgContent);
     if (disForm) disForm->GetDisabled(&isDisabled);
+    clickableForm = PR_TRUE;
+    break;
   }
   case NS_FORM_OPTGROUP:
   {
@@ -497,28 +504,34 @@ MicrobEalContextMenuInfo::CheckDomHtmlNode(nsIDOMNode *aNode, PRBool &aHasImage)
       aHasImage = PR_TRUE;
     } else
       mCtx->type |= GTK_MICROB_EAL_CONTEXT_EMBED;
+    rv = NS_ERROR_FAILURE; // do not inspect parent nodes
   }
   if (uTag.LowerCaseEqualsLiteral("embed")) {
     mCtx->type |= GTK_MICROB_EAL_CONTEXT_EMBED;
+    rv = NS_ERROR_FAILURE; // do not inspect parent nodes
   }
   else if (uTag.LowerCaseEqualsLiteral("html")) {
-    nsCOMPtr<nsIDOMNodeList> nodeList;
-    rv = mCtxDocument->GetElementsByTagName (NS_LITERAL_STRING("body"),
-                                             getter_AddRefs (nodeList));
-    if (NS_SUCCEEDED (rv) && nodeList) {
-      nsCOMPtr<nsIDOMNode> bodyNode;
-      nodeList->Item (0, getter_AddRefs (bodyNode));
-      nsString cssurl;
-      rv = GetCSSBackground (bodyNode, cssurl);
-      if (NS_SUCCEEDED (rv)) {
-        nsCString bgimg;
-        rv = ResolveBaseURL (cssurl, bgimg);
-        if (NS_FAILED (rv)) return NS_ERROR_FAILURE;
-        SetStringProperty ("image", bgimg.get());
-        mCtx->type |= GTK_MICROB_EAL_CONTEXT_IMAGE;
-        aHasImage = PR_TRUE;
+    if(!aHasImage) { // do not overwrite image url if already set
+      nsCOMPtr<nsIDOMNodeList> nodeList;
+      rv = mCtxDocument->GetElementsByTagName (NS_LITERAL_STRING("body"),
+                                               getter_AddRefs (nodeList));
+      if (NS_SUCCEEDED (rv) && nodeList) {
+        nsCOMPtr<nsIDOMNode> bodyNode;
+        nodeList->Item (0, getter_AddRefs (bodyNode));
+        nsString cssurl;
+        rv = GetCSSBackground (bodyNode, cssurl);
+        if (NS_SUCCEEDED (rv)) {
+          nsCString bgimg;
+          rv = ResolveBaseURL (cssurl, bgimg);
+          if (NS_FAILED (rv)) return NS_ERROR_FAILURE;
+          SetStringProperty ("image", bgimg.get());
+          mCtx->type |= GTK_MICROB_EAL_CONTEXT_IMAGE;
+          mCtx->type |= GTK_MICROB_EAL_CONTEXT_BGIMAGE;
+          aHasImage = PR_TRUE;
+        }
       }
     }
+    rv = NS_ERROR_FAILURE; // do not inspect parent nodes
   }
   else if (uTag.LowerCaseEqualsLiteral("a")) {
 
@@ -560,6 +573,7 @@ MicrobEalContextMenuInfo::CheckDomHtmlNode(nsIDOMNode *aNode, PRBool &aHasImage)
         }
       }
     }
+    rv = NS_ERROR_FAILURE; // do not inspect parent nodes
   }
   else if (uTag.LowerCaseEqualsLiteral("option") || uTag.LowerCaseEqualsLiteral("optgroup")) {
     mCtx->type |= GTK_MICROB_EAL_CONTEXT_NONE;
@@ -589,11 +603,8 @@ MicrobEalContextMenuInfo::CheckDomHtmlNode(nsIDOMNode *aNode, PRBool &aHasImage)
       SetStringProperty ("image", tmp);
       aHasImage = PR_TRUE;
     }
+    rv = NS_OK;
   }
-  else if (uTag.LowerCaseEqualsLiteral("span"))
-    rv = NS_OK;   // Just do nothing and proceed to the parent
-  else
-    rv = NS_ERROR_FAILURE;
 
   if (!aHasImage)
   {
@@ -605,6 +616,7 @@ MicrobEalContextMenuInfo::CheckDomHtmlNode(nsIDOMNode *aNode, PRBool &aHasImage)
       {
         SetStringProperty ("image", bgimg.get());
         mCtx->type |= GTK_MICROB_EAL_CONTEXT_IMAGE;
+        mCtx->type |= GTK_MICROB_EAL_CONTEXT_BGIMAGE;
         aHasImage = PR_TRUE;
       }
     }
@@ -727,10 +739,9 @@ MicrobEalContextMenuInfo::CheckDomImageElement(nsIDOMNode *node, nsString& aHref
   nsCOMPtr<imgIContainer> imgCont;
   request->GetImage(getter_AddRefs(imgCont));
   NS_ENSURE_TRUE(imgCont, rv);
-  nsCOMPtr<gfxIImageFrame> currentFrame;
-  imgCont->GetCurrentFrame(getter_AddRefs(currentFrame));
-  NS_ENSURE_TRUE(currentFrame, rv);
-  currentFrame->GetImageDataLength(aSize);
+  PRUint32 currentFrameIndex = 0;
+  imgCont->GetCurrentFrameIndex(&currentFrameIndex);
+  rv = imgCont->GetFrameImageDataLength(currentFrameIndex, aSize);
   return rv;
 }
 
@@ -821,7 +832,7 @@ MicrobEalContextMenuInfo::InsertTextToNode(nsIDOMNode *aDOMNode, nsAString &aStr
     NS_ENSURE_SUCCESS(rv, rv);
     nseditable->SetUserInput(buffer);
 
-    rv = DoHandleEventWithTarget(aDOMNode);
+    rv = SendFormInputEvent(targetNode);
     NS_ENSURE_SUCCESS(rv, rv);
    
     int len = aString.Length() + surr1;
@@ -864,9 +875,6 @@ MicrobEalContextMenuInfo::InsertTextToNode(nsIDOMNode *aDOMNode, nsAString &aStr
     } else
       buffer.Assign(aString);
 
-    input->SetValue(buffer);
-    nsinput->SetSelectionRange(selectionStart + surr1 + aString.Length(), selectionStart + surr1 + aString.Length());
-
     nsCOMPtr<nsIDOMNSEditableElement> nseditable;
     nseditable = do_QueryInterface(targetNode, &rv);
     NS_ENSURE_SUCCESS(rv, rv);
@@ -875,6 +883,16 @@ MicrobEalContextMenuInfo::InsertTextToNode(nsIDOMNode *aDOMNode, nsAString &aStr
     nseditable->GetEditor(getter_AddRefs(theEditor));
     NS_ENSURE_TRUE(theEditor, NS_OK);
 
+    if (mCtx->type & GTK_MICROB_EAL_CONTEXT_IPASSWORD)
+      theEditor->SetPasswordPosition(selectionStart+1);
+    else
+      theEditor->SetPasswordPosition(0);
+
+    nseditable->SetUserInput(buffer);
+    rv = SendFormInputEvent(targetNode);
+    NS_ENSURE_SUCCESS(rv, rv);
+    nsinput->SetSelectionRange(selectionStart + surr1 + aString.Length(), selectionStart + surr1 + aString.Length());
+
     nsCOMPtr<nsISelectionController> selectionController;
     theEditor->GetSelectionController(getter_AddRefs(selectionController));
     PRBool sync = PR_TRUE;
@@ -911,11 +929,12 @@ MicrobEalContextMenuInfo::InsertTextToNode(nsIDOMNode *aDOMNode, nsAString &aStr
 }
 
 nsresult
-MicrobEalContextMenuInfo::GetEventData()
+MicrobEalContextMenuInfo::GetEventData(bool& aIsMouseEvent)
 {
   nsresult rv;
 
   nsCOMPtr<nsIDOMMouseEvent> mouseEvent(do_QueryInterface(mCtxEvent, &rv));
+  aIsMouseEvent = (mouseEvent != nsnull);
   if (!mouseEvent) {
     nsCOMPtr<nsIDOMNSElement> nselement;
     nselement = do_QueryInterface(mOrigNode, &rv);
@@ -963,13 +982,22 @@ MicrobEalContextMenuInfo::GetEventData()
   return NS_OK;
 }
 
+nsresult
+MicrobEalContextMenuInfo::RecalculateContextData(nsIDOMNode* aDOMNode)
+{
+  NS_ENSURE_TRUE(mCtxEvent && aDOMNode, NS_ERROR_FAILURE);
+
+  UpdateContextDataFromNode(aDOMNode, true);
+
+  return NS_OK;
+}
 
 nsresult
 MicrobEalContextMenuInfo::UpdateContextDataFromElement(nsIDOMNSElement* aDOMElement)
 {
   nsCOMPtr<nsIDOMNode> eventNode = do_QueryInterface(aDOMElement);
 
-  return UpdateContextDataFromNode(eventNode.get());
+  return UpdateContextDataFromNode(eventNode.get(), false);
 }
 
 nsresult
@@ -1001,17 +1029,17 @@ MicrobEalContextMenuInfo::UpdateContextData(nsIDOMEvent *aDOMEvent)
   nsCOMPtr<nsIDOMNode> eventNode = do_QueryInterface(eventTarget, &rv);
   NS_ENSURE_SUCCESS(rv , NS_OK);
 
-  GetEventData();
-  return UpdateContextDataFromNode(eventNode);
+  bool isMouseEvent = false;
+  GetEventData(isMouseEvent);
+  return UpdateContextDataFromNode(eventNode, false, isMouseEvent);
 }
 
-
 nsresult
-MicrobEalContextMenuInfo::UpdateContextDataFromNode(nsIDOMNode *aDOMNode)
+MicrobEalContextMenuInfo::UpdateContextDataFromNode(nsIDOMNode *aDOMNode, bool aForce, bool aUpdateTitle)
 {
   nsresult rv;
 
-  if (aDOMNode == mOrigNode)
+  if (aDOMNode == mOrigNode && !aForce)
     return NS_OK;
   mEventNode = mOrigNode = aDOMNode;
 
@@ -1019,7 +1047,6 @@ MicrobEalContextMenuInfo::UpdateContextDataFromNode(nsIDOMNode *aDOMNode)
   mCtx->docindex = -1;
   mCtx->type = GTK_MICROB_EAL_CONTEXT_NONE;
 
-
   if (IsXULNode(mOrigNode, &mIsScrollableElement)) {
     mCtx->type |= GTK_MICROB_EAL_CONTEXT_XUL;
     IsXULEdit(mEventNode, &mCtx->type);
@@ -1080,20 +1107,25 @@ MicrobEalContextMenuInfo::UpdateContextDataFromNode(nsIDOMNode *aDOMNode)
   NS_ENSURE_SUCCESS(rv, rv);
 
   PRBool has_image = PR_FALSE;
-  while (node) {
-    if (NS_FAILED(CheckDomHtmlNode(node, has_image)))
-      break;
-
-    node->GetParentNode(getter_AddRefs(parentNode));
-    node = parentNode;
+  if (!(mCtx->type & GTK_MICROB_EAL_CONTEXT_FORM)) { // form controls do not have CSM
+    while (node) {
+      if (NS_FAILED(CheckDomHtmlNode(node, has_image)))
+        break;
+    
+      node->GetParentNode(getter_AddRefs(parentNode));
+      node = parentNode;
+    }
+    mCtx->type |= GTK_MICROB_EAL_CONTEXT_DOCUMENT;
   }
-  mCtx->type |= GTK_MICROB_EAL_CONTEXT_DOCUMENT;
 
-  nsString elementTitle;
-  if (NS_SUCCEEDED(GetElementTitle(aDOMNode, elementTitle)))
-    SetStringProperty("title", elementTitle);
-  else
-    SetStringProperty("title", "");
+  // update tooltip title only from mouse events
+  if (aUpdateTitle) {
+    nsString elementTitle;
+    if (NS_SUCCEEDED(GetElementTitle(aDOMNode, elementTitle)))
+      SetStringProperty("title", elementTitle);
+    else
+      SetStringProperty("title", "");
+  }
 
   return NS_OK;
 }
@@ -1237,11 +1269,15 @@ MicrobEalContextMenuInfo::GetElementTitle(nsIDOMNode *aNode, nsString &title)
 #include "nsIScrollableFrame.h"
 
 static inline nsIFrame* GetFrameFor(nsIView *aView)
-  { return static_cast<nsIFrame*>(aView->GetClientData()); }
+{ 
+  NS_ENSURE_TRUE(aView, nsnull);
+  return static_cast<nsIFrame*>(aView->GetClientData());
+}
 
 static inline nsIScrollableFrame*
 GetScrollableFrameFor(nsIScrollableView *aScrollableView)
 {
+  NS_ENSURE_TRUE(aScrollableView, nsnull);
   nsIFrame *frame = GetFrameFor(aScrollableView->View()->GetParent());
   nsIScrollableFrame *sf = do_QueryFrame(frame);
   return sf;
@@ -1250,10 +1286,13 @@ GetScrollableFrameFor(nsIScrollableView *aScrollableView)
 nsPresContext::ScrollbarStyles
 ScrollbarStylesOfView(nsIScrollableView *aScrollableView)
 {
-  nsIScrollableFrame *sf = GetScrollableFrameFor(aScrollableView);
-  return sf ? sf->GetScrollbarStyles() :
-              nsPresContext::ScrollbarStyles(NS_STYLE_OVERFLOW_HIDDEN,
-                                             NS_STYLE_OVERFLOW_HIDDEN);
+  if (aScrollableView) {
+    nsIScrollableFrame *sf = GetScrollableFrameFor(aScrollableView);
+    if (sf)
+        return sf->GetScrollbarStyles();
+  }
+  return nsPresContext::ScrollbarStyles(NS_STYLE_OVERFLOW_HIDDEN,
+                                        NS_STYLE_OVERFLOW_HIDDEN);
 }
 
 static nsIScrollableView*
@@ -1321,7 +1360,14 @@ MicrobEalContextMenuInfo::GetViewToScroll(nsIDOMNode *aDOMNode, nsIScrollableVie
   if (startFrame) {
     nsIScrollableViewProvider* svp = do_QueryFrame(startFrame);
     nsIScrollableView* sv;
-    nsIView* startView = svp && (sv = svp->GetScrollableView()) ? sv->View() : startFrame->GetAncestorWithViewExternal()->GetViewExternal();
+    nsIView* startView = NULL;
+    if (svp && (sv = svp->GetScrollableView()))
+      startView = sv->View();
+    else {
+      nsIFrame* ancView = startFrame->GetAncestorWithViewExternal();
+      if (ancView)
+        startView = ancView->GetViewExternal();
+    }
     NS_ASSERTION(startView, "No view to start searching for scrollable view from");
     scrollView = GetNearestScrollingView(startView, aDirection);
   }
@@ -1370,7 +1416,7 @@ MicrobEalContextMenuInfo::GetScrollableType(nsIScrollableView *aScrollView, PRBo
 }
 
 nsresult
-MicrobEalContextMenuInfo::ScrollScrollableView(PRInt32 aDX, PRInt32 aDY)
+MicrobEalContextMenuInfo::ScrollScrollableView(PRInt32 aDX, PRInt32 aDY, PRInt32 *aOverflowX, PRInt32 *aOverflowY)
 {
   int type = GetScrollableType();
   NS_ENSURE_TRUE(mScrollableView, NS_ERROR_FAILURE);
@@ -1395,11 +1441,14 @@ MicrobEalContextMenuInfo::ScrollScrollableView(PRInt32 aDX, PRInt32 aDY)
       mScrollableView = temp;
   }
   PRInt32 overflowX = 0, overflowY = 0;
-  return mScrollableView->ScrollByPixels(-aDX, -aDY, overflowX, overflowY);
+  nsresult rv = mScrollableView->ScrollByPixels(-aDX, -aDY, overflowX, overflowY);
+  if (aOverflowX) *aOverflowX = overflowX;
+  if (aOverflowY) *aOverflowY = overflowY;
+  return rv;
 }
 
 nsresult
-MicrobEalContextMenuInfo::DoHandleEventWithTarget(nsIDOMNode *aDOMNode)
+MicrobEalContextMenuInfo::SendFormInputEvent(nsIDOMNode *aDOMNode)
 {
   ULOG();
 
@@ -1418,20 +1467,8 @@ MicrobEalContextMenuInfo::DoHandleEventWithTarget(nsIDOMNode *aDOMNode)
   nsIPresShell *presShell = doc->GetPrimaryShell();
   NS_ENSURE_TRUE(presShell, NS_ERROR_FAILURE);
 
-  nsCOMPtr<nsIDOMHTMLInputElement> input;
-  input = do_QueryInterface(aDOMNode);
-
-  PRUint32 eventType = NS_FORM_INPUT;
-
-  if (input) {
-    nsString value;
-    input->GetValue(value);
-    if (!value.IsEmpty())
-      eventType = NS_FORM_CHANGE;
-  }
-
   nsEventStatus status = nsEventStatus_eIgnore;
-  nsInputEvent event(PR_TRUE, eventType, nsnull);
+  nsInputEvent event(PR_TRUE, NS_FORM_INPUT, nsnull);
   return presShell->HandleEventWithTarget(&event, nsnull, content, &status);
 }
 
@@ -1442,7 +1479,7 @@ MicrobEalContextMenuInfo::UpdateScrollableView(nsIDOMNode *aNode, int aDirection
 }
 
 nsresult
-MicrobEalContextMenuInfo::GetDomWindowScrollInfo(nsIDOMWindow *aDOMWindow, PRInt32 *aW, PRInt32 *aH, PRInt32 *aScrMX, PRInt32 *aScrMY, PRInt32 *aScrX, PRInt32 *aScrY, PRBool aFlush)
+MicrobEalContextMenuInfo::GetDomWindowScrollInfo(nsIDOMWindow *aDOMWindow, PRInt32 *aW, PRInt32 *aH, PRInt32 *aScrMX, PRInt32 *aScrMY, PRInt32 *aScrX, PRInt32 *aScrY, PRInt32 *aBidiOffset)
 {
   NS_ENSURE_ARG_POINTER(aDOMWindow);
   nsresult rv;
@@ -1451,8 +1488,31 @@ MicrobEalContextMenuInfo::GetDomWindowScrollInfo(nsIDOMWindow *aDOMWindow, PRInt
   rv = aDOMWindow->GetDocument(getter_AddRefs(domDoc));
   doc = do_QueryInterface(domDoc, &rv);
   NS_ENSURE_TRUE(doc, NS_ERROR_FAILURE);
+
   nsIPresShell *shell = doc->GetPrimaryShell();
   NS_ENSURE_TRUE(shell, NS_ERROR_FAILURE);
+
+  PRBool bidi = PR_FALSE;
+  if (doc->GetBidiEnabled()) {
+    // Internal API usage, should be implemented as part of CSSOM
+    nsIFrame *frame = nsnull;
+    nsIContent *root = doc->GetRootContent();
+    nsCOMPtr<nsIHTMLDocument> htmlDoc = do_QueryInterface(doc);
+    if (htmlDoc) {
+      nsIContent *bodyContent = htmlDoc->GetBodyContentExternal();
+      if (bodyContent)
+        root = bodyContent; // we can trust the document to hold on to it
+    }
+
+    if (root) {
+      nsIFrame *rootsFrame = shell->GetPrimaryFrameFor(root);
+      if (rootsFrame)
+        frame = rootsFrame;
+    }
+    if (frame)
+      bidi = frame->GetStyleVisibility()->mDirection == NS_STYLE_DIRECTION_RTL;
+  }
+
   nsCOMPtr<nsIViewManager> viewManager;
   viewManager = shell->GetViewManager();
   NS_ENSURE_TRUE(viewManager, NS_ERROR_FAILURE);
@@ -1467,19 +1527,6 @@ MicrobEalContextMenuInfo::GetDomWindowScrollInfo(nsIDOMWindow *aDOMWindow, PRInt
       rv |= mainWinInt->GetInnerWidth(aW);
     if (aH)
       rv |= mainWinInt->GetInnerHeight(aH);
-    if (!aFlush)
-      return rv;
-
-    if (aScrMX)
-      rv |= mainWinInt->GetScrollMaxX(aScrMX);
-    if (aScrMY)
-      rv |= mainWinInt->GetScrollMaxY(aScrMY);
-
-    if (aScrX)
-      rv |= aDOMWindow->GetScrollX(aScrX);
-    if (aScrY)
-      rv |= aDOMWindow->GetScrollY(aScrY);
-
     return rv;
   }
 
@@ -1489,26 +1536,14 @@ MicrobEalContextMenuInfo::GetDomWindowScrollInfo(nsIDOMWindow *aDOMWindow, PRInt
   if (aH)
     *aH = (PRInt32)floor(nsPresContext::AppUnitsToFloatCSSPixels(portRect.height));
 
-  if (aFlush) {
-    nsCOMPtr<nsIDOMWindowInternal> mainWinInt = do_QueryInterface(aDOMWindow, &rv);
-    NS_ENSURE_SUCCESS(rv, rv); // we should have internal aDOMWindow
-    if (aScrMX)
-      rv |= mainWinInt->GetScrollMaxX(aScrMX);
-    if (aScrMY)
-      rv |= mainWinInt->GetScrollMaxY(aScrMY);
-
-    if (aScrX)
-      rv |= aDOMWindow->GetScrollX(aScrX);
-    if (aScrY)
-      rv |= aDOMWindow->GetScrollY(aScrY);
-    return NS_OK;
-  }
-
   rv = scrollView->GetContainerSize(&scrolledSize.width, &scrolledSize.height);
   NS_ENSURE_SUCCESS(rv, rv);
 
+  PRInt32 scrMax = PR_MAX(0, (PRInt32)floor(nsPresContext::AppUnitsToFloatCSSPixels(scrolledSize.width - portRect.width)));
   if (aScrMX)
-    *aScrMX = PR_MAX(0, (PRInt32)floor(nsPresContext::AppUnitsToFloatCSSPixels(scrolledSize.width - portRect.width)));
+    *aScrMX = scrMax;
+  if (aBidiOffset && bidi)
+    *aBidiOffset = scrMax;
   if (aScrMY)
     *aScrMY = PR_MAX(0, (PRInt32)floor(nsPresContext::AppUnitsToFloatCSSPixels(scrolledSize.height - portRect.height)));
 
@@ -1517,8 +1552,12 @@ MicrobEalContextMenuInfo::GetDomWindowScrollInfo(nsIDOMWindow *aDOMWindow, PRInt
   nscoord xPos, yPos;
   rv = scrollView->GetScrollPosition(xPos, yPos);
   NS_ENSURE_SUCCESS(rv, rv);
-  if (aScrX)
+  if (aScrX) {
     *aScrX = nsPresContext::AppUnitsToIntCSSPixels(xPos);
+    if (bidi)
+      *aScrX += scrMax;
+  }
+
   if (aScrY)
     *aScrY = nsPresContext::AppUnitsToIntCSSPixels(yPos);
 
