@@ -33,6 +33,8 @@
 
 /* GMozillaHistoryList */
 
+static GMozillaHistoryList *singleton_instance = NULL;
+static guint ref_count = 0;
 
 int
 g_mozilla_history_list_clear_all ()
@@ -52,26 +54,28 @@ g_mozilla_history_list_position_ok (GWebHistoryList *self, gint pos)
 static void
 g_mozilla_history_list_delete_item (GMozillaHistoryList *self, gint pos)
 {
+    g_return_if_fail(G_IS_MOZILLA_HISTORY_LIST(self));
+
     guint rv;
     TRACE_LOG();
 
     if (!g_mozilla_history_list_position_ok((GWebHistoryList*)self, pos)) {
         return;
     }
-    gchar *string = (char*)(G_MOZILLA_HISTORY_LIST(self)->items)[pos].url;
-    if (!string) {
+    GtkWebHistoryItem *item = (GtkWebHistoryItem *)g_list_nth_data(self->items, pos);
+    if (!item->url)
         return ;
-    }
-    rv = g_mozilla_cpp_history_list_remove(string);
+    rv = g_mozilla_cpp_history_list_remove(item->url);
     /* we don't have NS_FAILED, this is NS_FAILED */
     if (rv & 0x80000000) {
         TRACE_LOG("gtk_moz_embed_common_remove_history failed");
-        return ;
+        return;
     }
 
-    EAL_IF_GFREE(string);
-    string = (char*)(G_MOZILLA_HISTORY_LIST(self)->items)[pos].title;
-    EAL_IF_GFREE(string);
+    self->items = g_list_remove(self->items, item);
+    EAL_GFREE(item->url);
+    EAL_IF_GFREE(item->title);
+    EAL_GFREE(item);
 }
 
 static GObject*
@@ -86,15 +90,19 @@ static gint
 g_mozilla_history_list_get_items_count (GMozillaHistoryList *self)
 {
     TRACE_LOG();
-    return G_MOZILLA_HISTORY_LIST(self)->num_items;
+    g_return_val_if_fail(G_IS_MOZILLA_HISTORY_LIST(self), 0);
+    return self->num_items;
 }
 
 static const gchar*
 g_mozilla_history_list_get_item_title (GWebHistoryList *self, gint pos)
 {
     TRACE_LOG();
-    if (g_mozilla_history_list_position_ok(self, pos)) {
-        return (G_MOZILLA_HISTORY_LIST(self)->items)[pos].title;
+    if (G_IS_MOZILLA_HISTORY_LIST(self) &&
+        g_mozilla_history_list_position_ok(self, pos)) {
+        GtkWebHistoryItem *item = (GtkWebHistoryItem *)g_list_nth_data(
+            G_MOZILLA_HISTORY_LIST(self)->items, pos);
+        return item->title;
     }
     return NULL;
 }
@@ -103,8 +111,11 @@ static const gchar*
 g_mozilla_history_list_get_item_url (GWebHistoryList *self, gint pos)
 {
     TRACE_LOG();
-    if (g_mozilla_history_list_position_ok(self, pos)) {
-        return (G_MOZILLA_HISTORY_LIST(self)->items)[pos].url;
+    if (G_IS_MOZILLA_HISTORY_LIST(self) &&
+        g_mozilla_history_list_position_ok(self, pos)) {
+        GtkWebHistoryItem *item = (GtkWebHistoryItem *)g_list_nth_data(
+            G_MOZILLA_HISTORY_LIST(self)->items, pos);
+        return item->url;
     }
     return NULL;
 }
@@ -120,8 +131,11 @@ static GTime
 g_mozilla_history_list_get_item_time (GWebHistoryList *self, gint pos)
 {
     TRACE_LOG();
-    if (g_mozilla_history_list_position_ok(self, pos)) {
-	return (G_MOZILLA_HISTORY_LIST(self)->items)[pos].accessed;
+    if (G_IS_MOZILLA_HISTORY_LIST(self) &&
+        g_mozilla_history_list_position_ok(self, pos)) {
+        GtkWebHistoryItem *item = (GtkWebHistoryItem *)g_list_nth_data(
+            G_MOZILLA_HISTORY_LIST(self)->items, pos);
+        return item->accessed;
     }
     return 0;
 }
@@ -130,18 +144,19 @@ static void
 g_mozilla_history_list_destroy (GMozillaHistoryList *self)
 {
     TRACE_LOG();
-    gint index;
-    for (index = 0; index < self->num_items; index++) {
-        gchar *url = (gchar *) g_mozilla_history_list_get_item_url((GWebHistoryList *)self, index);
-        gchar *title = (gchar *) g_mozilla_history_list_get_item_title((GWebHistoryList *)self, index);
-        EAL_IF_GFREE(url);
-        EAL_IF_GFREE(title);
-#if 0
-        gchar *thumbnail_fn = (gchar *) g_mozilla_shistory_list_get_item_thumbnail((GWebHistoryList *)self, index);
-        EAL_IF_GFREE(thumbnail_fn);
-#endif
+
+    // Don't mess with cached array with window history lists
+    gint window_history = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(self), "window-history"));
+
+    if (ref_count && !window_history)
+        ref_count--;
+
+    if (ref_count == 0 || window_history) {
+        g_list_foreach(self->items, free_history_item, NULL);
+        EAL_GFREE_FUNC(self->items, g_list_free);
+        if (!window_history)
+            singleton_instance = NULL;
     }
-    g_free(G_MOZILLA_HISTORY_LIST(self)->items);
 }
 
 static void
@@ -212,42 +227,66 @@ GObject*
 g_mozilla_history_list_new (void)
 {
     TRACE_LOG();
-    GObject* instance = G_OBJECT(g_object_new(g_mozilla_history_list_get_type(), NULL));
-    G_MOZILLA_HISTORY_LIST(instance)->isSHistory = FALSE;
-    G_MOZILLA_HISTORY_LIST(instance)->num_items =
-                g_mozilla_cpp_get_history_list(
-                        &(G_MOZILLA_HISTORY_LIST(instance))->items);
-    return G_OBJECT(instance);
+
+    if (!singleton_instance) { 
+        singleton_instance = G_MOZILLA_HISTORY_LIST(g_object_new(g_mozilla_history_list_get_type(), NULL));
+        singleton_instance->isSHistory = FALSE;
+        singleton_instance->num_items = g_mozilla_cpp_get_history_list(
+                        &singleton_instance->items,
+                        singleton_instance->num_items);
+        ref_count = 1;
+    } else
+        ref_count++;
+
+    return G_OBJECT(singleton_instance);
 }
 
-typedef struct
+typedef struct _HistoryMessage HistoryMessage;
+struct _HistoryMessage
 {
-  gpointer engine;
-  unsigned int message;
-  int number;
-  GWebEngineEncoding encoding;
-} HistStruct;
+    GMozillaEngine *engine;
+    guint message;
+    gint number;
+    GWebEngineEncoding encoding;
+};
 
-//Workaround for bug #102760
-static gboolean send_history_message_delayed(gpointer data)
+static gboolean
+idle_cb (gpointer user_data)
 {
-  HistStruct* hist_data = data;
-  g_return_val_if_fail(hist_data && hist_data->engine, FALSE);
-  g_signal_emit_by_name (G_OBJECT(hist_data->engine), G_WEBWIDGET_SIGNAL_HISTORY_CHANGED, hist_data->message, hist_data->number, hist_data->encoding);
-  free(hist_data);
-  return FALSE;
+    TRACE_LOG();
+
+    HistoryMessage *msg = (HistoryMessage *)user_data;
+
+    if (singleton_instance)
+        singleton_instance->num_items =
+            g_mozilla_cpp_get_history_list(&singleton_instance->items,
+                                           singleton_instance->num_items);
+
+    g_object_set_data(G_OBJECT(msg->engine), "async", GINT_TO_POINTER(1));
+    g_signal_emit_by_name(msg->engine,
+                          G_WEBWIDGET_SIGNAL_HISTORY_CHANGED,
+                          msg->message, msg->number, msg->encoding);
+
+    g_free(msg);
+    return FALSE;
 }
 
 void send_history_message(GMozillaEngine* engine, unsigned int message, int number, GWebEngineEncoding encoding)
 {
-  HistStruct* hist_data = malloc(sizeof(HistStruct));
-  g_return_if_fail(hist_data);
-  hist_data->engine = engine;
-  hist_data->message = message;
-  hist_data->number = number;
-  hist_data->encoding = encoding;
-  //send message in next loop iteration
-  g_idle_add((GSourceFunc)send_history_message_delayed, (gpointer)hist_data);
+  TRACE_LOG();
+
+  HistoryMessage *msg = g_try_new0(HistoryMessage, 1);
+  if (!msg)
+      return;
+
+  msg->engine = engine;
+  msg->message = message;
+  msg->number = number;
+  msg->encoding = encoding;
+
+  // Let engine side process to handle history addition before
+  // sending the notification forward
+  g_idle_add ((GSourceFunc)idle_cb, msg);
 }
 
 
