@@ -185,8 +185,6 @@ g_mozilla_web_destroy (GMozillaWeb *self)
 
     if (self->shutdown_xpcom) {
       g_mozilla_plugins_enable(self, FALSE);
-      g_web_bus_unref(self->web_bus);
-      self->web_bus = NULL;
       gtk_main_quit();
     }
 }
